@@ -1,4 +1,5 @@
-import * as path from 'path'
+import * as Process from 'child_process'
+import * as Path from 'path'
 import * as TurndownService from 'turndown'
 import * as LSP from 'vscode-languageserver'
 import { TextDocument } from 'vscode-languageserver-textdocument'
@@ -7,12 +8,15 @@ import Analyzer from './analyser'
 import * as Builtins from './builtins'
 import * as config from './config'
 import Executables from './executables'
+import Linter from './linter'
 import { initializeParser } from './parser'
 import * as ReservedWords from './reservedWords'
 import { BashCompletionItem, CompletionItemDataType } from './types'
 import { uniqueBasedOnHash } from './util/array'
 import { getShellDocumentation } from './util/sh'
 
+const PARAMETER_EXPANSION_PREFIXES = new Set(['$', '${'])
+
 /**
  * The BashServer glues together the separate components to implement
  * the various parts of the Language Server Protocol.
@@ -24,7 +28,7 @@ export default class BashServer {
    */
   public static async initialize(
     connection: LSP.Connection,
-    { rootPath }: LSP.InitializeParams,
+    { rootPath, capabilities }: LSP.InitializeParams,
   ): Promise<BashServer> {
     const parser = await initializeParser()
 
@@ -37,27 +41,32 @@ export default class BashServer {
     return Promise.all([
       Executables.fromPath(PATH),
       Analyzer.fromRoot({ connection, rootPath, parser }),
-    ]).then(xs => {
-      const executables = xs[0]
-      const analyzer = xs[1]
-      return new BashServer(connection, executables, analyzer)
+      new Linter({ executablePath: config.getShellcheckPath() }),
+    ]).then(([executables, analyzer, linter]) => {
+      return new BashServer(connection, executables, analyzer, linter, capabilities)
     })
   }
 
   private executables: Executables
   private analyzer: Analyzer
+  private linter: Linter
 
   private documents: LSP.TextDocuments<TextDocument> = new LSP.TextDocuments(TextDocument)
   private connection: LSP.Connection
+  private clientCapabilities: LSP.ClientCapabilities
 
   private constructor(
     connection: LSP.Connection,
     executables: Executables,
     analyzer: Analyzer,
+    linter: Linter,
+    capabilities: LSP.ClientCapabilities,
   ) {
     this.connection = connection
     this.executables = executables
     this.analyzer = analyzer
+    this.linter = linter
+    this.clientCapabilities = capabilities
   }
 
   /**
@@ -68,15 +77,28 @@ export default class BashServer {
     // The content of a text document has changed. This event is emitted
     // when the text document first opened or when its content has changed.
     this.documents.listen(this.connection)
-    this.documents.onDidChangeContent(change => {
+    this.documents.onDidChangeContent(async (change) => {
       const { uri } = change.document
-      const diagnostics = this.analyzer.analyze(uri, change.document)
+
+      // Load the tree for the modified contents into the analyzer:
+      const analyzeDiagnostics = this.analyzer.analyze(uri, change.document)
+
+      // Run shellcheck diagnostics:
+      let diagnostics: LSP.Diagnostic[] = []
+
+      const folders = this.clientCapabilities.workspace?.workspaceFolders
+        ? await connection.workspace.getWorkspaceFolders()
+        : []
+      const lintDiagnostics = await this.linter.lint(change.document, folders || [])
+      diagnostics = diagnostics.concat(lintDiagnostics)
+
+      // Treesitter's diagnostics can be a bit inaccurate, so we only merge the
+      // analyzer's diagnostics if the setting is enabled:
       if (config.getHighlightParsingError()) {
-        connection.sendDiagnostics({
-          uri: change.document.uri,
-          diagnostics,
-        })
+        diagnostics = diagnostics.concat(analyzeDiagnostics)
       }
+
+      connection.sendDiagnostics({ uri, diagnostics })
     })
 
     // Register all the handlers for the LSP events.
@@ -88,6 +110,8 @@ export default class BashServer {
     connection.onReferences(this.onReferences.bind(this))
     connection.onCompletion(this.onCompletion.bind(this))
     connection.onCompletionResolve(this.onCompletionResolve.bind(this))
+
+    // FIXME: re-lint on workspace folder change
   }
 
   /**
@@ -100,6 +124,7 @@ export default class BashServer {
       textDocumentSync: LSP.TextDocumentSyncKind.Full,
       completionProvider: {
         resolveProvider: true,
+        triggerCharacters: ['$', '{'],
       },
       hoverProvider: true,
       documentHighlightProvider: true,
@@ -120,6 +145,16 @@ export default class BashServer {
     )
   }
 
+  private getCommandNameAtPoint(
+    params: LSP.ReferenceParams | LSP.TextDocumentPositionParams,
+  ): string | null {
+    return this.analyzer.commandNameAtPoint(
+      params.textDocument.uri,
+      params.position.line,
+      params.position.character,
+    )
+  }
+
   private logRequest({
     request,
     params,
@@ -135,6 +170,53 @@ export default class BashServer {
     )
   }
 
+  private getDocumentationForSymbol({
+    currentUri,
+    symbol,
+  }: {
+    symbol: LSP.SymbolInformation
+    currentUri: string
+  }): string {
+    const symbolUri = symbol.location.uri
+    const symbolStarLine = symbol.location.range.start.line
+
+    const commentAboveSymbol = this.analyzer.commentsAbove(symbolUri, symbolStarLine)
+    const symbolDocumentation = commentAboveSymbol ? `\n\n${commentAboveSymbol}` : ''
+    const hoverHeader = `### ${symbolKindToDescription(symbol.kind)}: **${symbol.name}**`
+    const symbolLocation =
+      symbolUri !== currentUri
+        ? `in ${Path.relative(currentUri, symbolUri)}`
+        : `on line ${symbolStarLine + 1}`
+
+    return `${hoverHeader} - *defined ${symbolLocation}*${symbolDocumentation}`
+  }
+
+  private getCompletionItemsForSymbols({
+    symbols,
+    currentUri,
+  }: {
+    symbols: LSP.SymbolInformation[]
+    currentUri: string
+  }): BashCompletionItem[] {
+    return deduplicateSymbols({ symbols, currentUri }).map(
+      (symbol: LSP.SymbolInformation) => ({
+        label: symbol.name,
+        kind: symbolKindToCompletionKind(symbol.kind),
+        data: {
+          name: symbol.name,
+          type: CompletionItemDataType.Symbol,
+        },
+        documentation:
+          symbol.location.uri !== currentUri
+            ? this.getDocumentationForSymbol({
+                currentUri,
+                symbol,
+              })
+            : undefined,
+      }),
+    )
+  }
+
   private async onHover(
     params: LSP.TextDocumentPositionParams,
   ): Promise<LSP.Hover | null> {
@@ -150,22 +232,29 @@ export default class BashServer {
     const explainshellEndpoint = config.getExplainshellEndpoint()
     if (explainshellEndpoint) {
       this.connection.console.log(`Query ${explainshellEndpoint}`)
-      const response = await this.analyzer.getExplainshellDocumentation({
-        params,
-        endpoint: explainshellEndpoint,
-      })
+      try {
+        const response = await this.analyzer.getExplainshellDocumentation({
+          params,
+          endpoint: explainshellEndpoint,
+        })
 
-      if (response.status === 'error') {
-        this.connection.console.log(
-          `getExplainshellDocumentation returned: ${JSON.stringify(response, null, 4)}`,
-        )
-      } else {
-        return {
-          contents: {
-            kind: 'markdown',
-            value: new TurndownService().turndown(response.helpHTML),
-          },
+        if (response.status === 'error') {
+          this.connection.console.log(
+            `getExplainshellDocumentation returned: ${JSON.stringify(response, null, 4)}`,
+          )
+        } else {
+          return {
+            contents: {
+              kind: 'markdown',
+              value: new TurndownService().turndown(response.helpHTML),
+            },
+          }
         }
+      } catch (error) {
+        const errorMessage = error instanceof Error ? error.message : error
+        this.connection.console.warn(
+          `getExplainshellDocumentation exception: ${errorMessage}`,
+        )
       }
     }
 
@@ -188,15 +277,9 @@ export default class BashServer {
         currentUri,
       })
         // do not return hover referencing for the current line
-        .filter(symbol => symbol.location.range.start.line !== params.position.line)
+        .filter((symbol) => symbol.location.range.start.line !== params.position.line)
         .map((symbol: LSP.SymbolInformation) =>
-          symbol.location.uri !== currentUri
-            ? `${symbolKindToDescription(symbol.kind)} defined in ${path.relative(
-                currentUri,
-                symbol.location.uri,
-              )}`
-            : `${symbolKindToDescription(symbol.kind)} defined on line ${symbol.location
-                .range.start.line + 1}`,
+          this.getDocumentationForSymbol({ currentUri, symbol }),
         )
 
       if (symbolDocumentation.length === 1) {
@@ -231,12 +314,14 @@ export default class BashServer {
   ): LSP.DocumentHighlight[] | null {
     const word = this.getWordAtPoint(params)
     this.logRequest({ request: 'onDocumentHighlight', params, word })
+
     if (!word) {
-      return null
+      return []
     }
+
     return this.analyzer
       .findOccurrences(params.textDocument.uri, word)
-      .map(n => ({ range: n.range }))
+      .map((n) => ({ range: n.range }))
   }
 
   private onReferences(params: LSP.ReferenceParams): LSP.Location[] | null {
@@ -249,23 +334,74 @@ export default class BashServer {
   }
 
   private onCompletion(params: LSP.TextDocumentPositionParams): BashCompletionItem[] {
-    const word = this.getWordAtPoint(params)
+    const word = this.getWordAtPoint({
+      ...params,
+      position: {
+        line: params.position.line,
+        // Go one character back to get completion on the current word
+        character: Math.max(params.position.character - 1, 0),
+      },
+    })
     this.logRequest({ request: 'onCompletion', params, word })
 
+    if (word && word.startsWith('#')) {
+      // Inside a comment block
+      return []
+    }
+    if (word && word === '{') {
+      // We should not complete when it is not prefixed by a $.
+      // This case needs to be here
+      // because { is a completionProvider triggerCharacter.
+      return []
+    }
+
+    let options: string[] = []
+    if (word && word.startsWith('-')) {
+      const commandName = this.getCommandNameAtPoint({
+        ...params,
+        position: {
+          line: params.position.line,
+          // Go one character back to get completion on the current word
+          character: Math.max(params.position.character - 1, 0),
+        },
+      })
+
+      if (commandName) {
+        options = getCommandOptions(commandName, word)
+      }
+    }
+
     const currentUri = params.textDocument.uri
 
+    // TODO: an improvement here would be to detect if the current word is
+    // not only a parameter expansion prefix, but also if the word is actually
+    // inside a parameter expansion (e.g. auto completing on a word $MY_VARIA).
+    const shouldCompleteOnVariables = word
+      ? PARAMETER_EXPANSION_PREFIXES.has(word)
+      : false
+
     const symbolCompletions =
       word === null
         ? []
-        : getCompletionItemsForSymbols({
-            symbols: this.analyzer.findSymbolsMatchingWord({
-              exactMatch: false,
-              word,
-            }),
+        : this.getCompletionItemsForSymbols({
+            symbols: shouldCompleteOnVariables
+              ? this.analyzer.getAllVariableSymbols()
+              : this.analyzer.findSymbolsMatchingWord({
+                  exactMatch: false,
+                  word,
+                }),
             currentUri,
           })
 
-    const reservedWordsCompletions = ReservedWords.LIST.map(reservedWord => ({
+    if (shouldCompleteOnVariables) {
+      // In case we auto complete on a word that starts a parameter expansion,
+      // we do not return anything else than variable/parameter suggestions.
+      // Note: that LSP clients should not call onCompletion in the middle
+      // of a word, so the following should work for client.
+      return symbolCompletions
+    }
+
+    const reservedWordsCompletions = ReservedWords.LIST.map((reservedWord) => ({
       label: reservedWord,
       kind: LSP.SymbolKind.Interface, // ??
       data: {
@@ -276,8 +412,8 @@ export default class BashServer {
 
     const programCompletions = this.executables
       .list()
-      .filter(executable => !Builtins.isBuiltin(executable))
-      .map(executable => {
+      .filter((executable) => !Builtins.isBuiltin(executable))
+      .map((executable) => {
         return {
           label: executable,
           kind: LSP.SymbolKind.Function,
@@ -288,7 +424,7 @@ export default class BashServer {
         }
       })
 
-    const builtinsCompletions = Builtins.LIST.map(builtin => ({
+    const builtinsCompletions = Builtins.LIST.map((builtin) => ({
       label: builtin,
       kind: LSP.SymbolKind.Interface, // ??
       data: {
@@ -297,21 +433,26 @@ export default class BashServer {
       },
     }))
 
+    const optionsCompletions = options.map((option) => ({
+      label: option,
+      kind: LSP.SymbolKind.Interface,
+      data: {
+        name: option,
+        type: CompletionItemDataType.Symbol,
+      },
+    }))
+
     const allCompletions = [
       ...reservedWordsCompletions,
       ...symbolCompletions,
       ...programCompletions,
       ...builtinsCompletions,
+      ...optionsCompletions,
     ]
 
     if (word) {
-      if (word.startsWith('#')) {
-        // Inside a comment block
-        return []
-      }
-
       // Filter to only return suffixes of the current word
-      return allCompletions.filter(item => item.label.startsWith(word))
+      return allCompletions.filter((item) => item.label.startsWith(word))
     }
 
     return allCompletions
@@ -364,15 +505,15 @@ function deduplicateSymbols({
 
   const getSymbolId = ({ name, kind }: LSP.SymbolInformation) => `${name}${kind}`
 
-  const symbolsCurrentFile = symbols.filter(s => isCurrentFile(s))
+  const symbolsCurrentFile = symbols.filter((s) => isCurrentFile(s))
 
   const symbolsOtherFiles = symbols
-    .filter(s => !isCurrentFile(s))
+    .filter((s) => !isCurrentFile(s))
     // Remove identical symbols matching current file
     .filter(
-      symbolOtherFiles =>
+      (symbolOtherFiles) =>
         !symbolsCurrentFile.some(
-          symbolCurrentFile =>
+          (symbolCurrentFile) =>
             getSymbolId(symbolCurrentFile) === getSymbolId(symbolOtherFiles),
         ),
     )
@@ -380,32 +521,6 @@ function deduplicateSymbols({
   return uniqueBasedOnHash([...symbolsCurrentFile, ...symbolsOtherFiles], getSymbolId)
 }
 
-function getCompletionItemsForSymbols({
-  symbols,
-  currentUri,
-}: {
-  symbols: LSP.SymbolInformation[]
-  currentUri: string
-}): BashCompletionItem[] {
-  return deduplicateSymbols({ symbols, currentUri }).map(
-    (symbol: LSP.SymbolInformation) => ({
-      label: symbol.name,
-      kind: symbolKindToCompletionKind(symbol.kind),
-      data: {
-        name: symbol.name,
-        type: CompletionItemDataType.Symbol,
-      },
-      documentation:
-        symbol.location.uri !== currentUri
-          ? `${symbolKindToDescription(symbol.kind)} defined in ${path.relative(
-              currentUri,
-              symbol.location.uri,
-            )}`
-          : undefined,
-    }),
-  )
-}
-
 function symbolKindToCompletionKind(s: LSP.SymbolKind): LSP.CompletionItemKind {
   switch (s) {
     case LSP.SymbolKind.File:
@@ -474,3 +589,21 @@ const getMarkdownContent = (documentation: string): LSP.MarkupContent => ({
   // Passed as markdown for syntax highlighting
   kind: 'markdown' as const,
 })
+
+function getCommandOptions(name: string, word: string): string[] {
+  // TODO: The options could be cached.
+  const options = Process.spawnSync(Path.join(__dirname, '../src/get-options.sh'), [
+    name,
+    word,
+  ])
+
+  if (options.status !== 0) {
+    return []
+  }
+
+  return options.stdout
+    .toString()
+    .split('\t')
+    .map((l) => l.trim())
+    .filter((l) => l.length > 0)
+}
